# tessif.frused.defaults
"""
:mod:`~tessif.frused.defaults` is a :mod:`tessif` subpackage providing
frequently needed defaults. Everything loosely associated with fallback
values as well as filter and sorting templates is aggregated here.
"""
import tessif.frused.namedtuples as nts
# from matplotlib import pyplot as plt
import tessif.frused.spellings as spellings

nxgrph_node_shapes = {
    'bus': 'o',
    'connector': 'o',
    'commodity_source': 'o',
    'default_source': 'o',
    'sink': '8',
    'solar': 's',
    'storage': 's',
    'transformer': '8',
    'wind': 'h',
}
"""
:mod:`~tessif.visualize.nxgrph` node shape visualization defaults.

.. execute_code::
    :hide_code:
    :hide_headers:
    :hide_output:

    from tessif.frused.paths import doc_dir
    import os
    from tessif.frused.defaults import nxgrph_node_shapes
    import dutils

    df = dutils.to_dataframe(nxgrph_node_shapes, columns=2, fillvalue='')

    path = os.path.join(
        doc_dir, 'source', 'api', 'frused',
        'defaults', 'node_shapes.csv')

    df.to_csv(path, header=None, index=None, na_rep='None')

.. csv-table::
    :file: source/api/frused/defaults/node_shapes.csv
"""

dcgrph_node_shapes = {
    'default_source': 'round-rectangle',
    'commodity_source': 'round-rectangle',
    'solar': 'round-diamond',
    'wind': 'diamond',
    'bus': 'ellipse',
    'transformer': 'round-octagon',
    'connector': 'rectangle',

    'sink': 'ellipse',
    'storage': 'round-hexagon',
}
""":mod:`~tessif.visualize.dcgrph` node shape visualization defaults."""

nxgrph_visualize_defaults = {
    # node_defaults:
    'node_labels': None,
    'node_shape': 'o',
    'node_size': 3000,
    'node_minimum_size': 0.1 * 3000,
    'node_variable_size_scaling': 0.5,
    'node_fill_size': 3000,
    'node_color': '#AFAFAF',
    'node_color_map': ['#AFAFAF'],
    'node_alpha': 1.0,
    'node_font_size': 11,
    'node_font_weight': 'light',
    # edge defaults:
    'edge_labels': None,
    'edge_width': 1,
    'edge_color': 'black',
    'edge_arrowstyle': 'simple',
    'edge_arrowsize': 7,
    'edge_vmin': 0.0,
    'edge_vmax': 1.0,
    'edge_cmap': None,  # plt.cm.Greys,
    'edge_len': 1.0,
    'edge_minimum_grey': 0.15,
    'edge_minimum_weight': 0.1,
    'edge_minimum_width': 0.1,
    # legend defaults:
    'legend_labelspacing': 1,
    'legend_title': None,
    'legend_bbox_to_anchor': (1.0, 1),
    'legend_loc': 'best',
    'legend_borderaxespad': 0,
}
"""
:mod:`~tessif.visualize.nxgrph` drawing defaults.

Used as described in :func:`~tessif.visualize.nxgrph.draw_graph`'s
:paramref:`~tessif.visualize.nxgrph.draw_graph.kwargs`.

.. execute_code::
    :hide_code:
    :hide_headers:
    :hide_output:

    from tessif.frused.paths import doc_dir
    import os
    from tessif.frused.defaults import nxgrph_visualize_defaults
    import dutils

    nxgrph_visualize_defaults['edge_cmap'] = '.'.join(
        ['matplotlib.pyplot.cm', nxgrph_visualize_defaults['edge_cmap'].name])

    df = dutils.to_dataframe(
             nxgrph_visualize_defaults, columns=2, fillvalue='')

    path = os.path.join(
         doc_dir, 'source', 'api', 'frused',
        'defaults', 'nxgrph_visualize_defaults.csv')
    df.to_csv(path, header=None, index=None, na_rep='None')

.. csv-table::
    :file: source/api/frused/defaults/nxgrph_visualize_defaults.csv
"""

dcgrph_visualize_defaults = {
    # node_defaults:
    'node_labels': None,
    'node_label_position': 'center',
    'node_shape': "ellipse",
    'node_size': 90,
    # 'node_size_basic': 100,
    'node_color': '#AFAFAF',
    'node_font_size': 15,
    'node_minimum_size': 1,
    'node_variable_size': 40,
    'node_border_width': 0,
    'node_fill_border_width': 1.5,  # used when partially filling a node
    'node_border_style': None,  # 'solid',
    'node_border_color': "black",
    'node_font_weight': 550,
    # edge defaults:
    'edge_labels': None,
    'edge_width': 7,
    'edge_width_basic': 1,
    'edge_color': 'black',
    'edge_arrow_color': 'black',
    'edge_arrowsize': 1,
    'edge_arrowstyle': 'triangle',
    'edge_style': 'bezier',
    'edge_linestyle': 'solid',
    # replacing dotted by larger spaced dashes
    'edge_dot_repl_pattern': [6, 6],
    'edge_minimum_width': 0.5,
    'edge_minimum_grey': 0.15,
    'edge_minimum_weight': nxgrph_visualize_defaults['edge_minimum_weight'],
}
""":mod:`~tessif.visualize.dcgrph` drawing defaults.

Used as described in :func:`~tessif.visualize.dcgrph.draw_graph`'s
:paramref:`~tessif.visualize.dcgrph.draw_graph.kwargs`.
"""

nxgrph_visualize_tags = nts.AttributeGroupings('node_', 'edge_', 'legend_')
"""
:mod:`~tessif.visualize.nxgrph` attribute tags. Used as described in
:func:`~tessif.visualize.nxgrph.draw_graph`'s
:paramref:`~tessif.visualize.nxgrph.draw_graph.kwargs`. Using
:attr:`~tessif.frused.namedtuples.AttributeGroupings` for sub categorization
as well as easy to maintain and expand code.

Currently used tags:

.. execute_code::
    :hide_code:
    :hide_headers:
    :hide_output:

    from tessif.frused.paths import doc_dir
    import os
    from tessif.frused.defaults import nxgrph_visualize_tags as nvt
    import pandas

    path = os.path.join(
        doc_dir, 'source', 'api', 'frused',
        'defaults', 'nxgrph_visualize_tags.csv')

    df = pandas.DataFrame(data=nvt._asdict().values(),
                          index=nvt._asdict().keys())

    for row, value in enumerate(df[0]):
        df[0][row] = '{}{}{}'.format('``', value, '``')

    df.to_csv(path, header=None)

.. csv-table:: first column labels the namedtuples field
    :file: source/api/frused/defaults/nxgrph_visualize_tags.csv
    :stub-columns: 1
"""

nxgrph_visualize_xcptns = nts.AttributeGroupings(
    # node exceptions
    ['node_size', 'node_color', 'node_shape', 'node_fill_size'],
    # edge exceptions
    ['edge_labels', 'edge_color', 'edge_vmin', 'edge_vmax', 'edge_cmap'],
    # legend exceptions
    [])
"""
:mod:`~tessif.visualize.nxgrph` filter exceptions. Used as described in
:func:`~tessif.visualize.nxgrph.draw_graph`'s
:paramref:`~tessif.visualize.nxgrph.draw_graph.kwargs`. Using
:attr:`~tessif.frused.namedtuples.AttributeGroupings` for sub categorization
as well as easy to maintain and expand code.

Currently used exceptions:

.. execute_code::
    :hide_code:
    :hide_headers:
    :hide_output:

    from tessif.frused.paths import doc_dir
    import os
    from tessif.frused.defaults import nxgrph_visualize_xcptns as nvx
    import pandas

    path = os.path.join(
         doc_dir, 'source', 'api', 'frused',
        'defaults', 'nxgrph_visualize_xcptns.csv')

    df = pandas.DataFrame(data=nvx._asdict().values(),
                          index=nvx._asdict().keys())

    df.to_csv(path, header=None)

.. csv-table:: first column labels the namedtuples field
    :file: source/api/frused/defaults/nxgrph_visualize_xcptns.csv
    :stub-columns: 1
"""

nx_label_kwargs = {
    'nodes': [
        'ax', 'nodelist', 'node_size', 'node_fill_size', 'node_color',
        'node_shape', 'alpha', 'cmap', 'vmin', 'vmax', 'linewidts',
        'edgecolors', 'label'],
    'labels': [
        'labels', 'font_size', 'font_color',  'font_family', 'font_weight',
        'alpha', 'horizontalalignment', 'verticalalignment', 'ax'],
    'edges': [
        'edglist', 'width', 'edge_color', 'style', 'alpha', 'edge_cmap',
        'edge_vmin', 'edge_vmax', 'ax', 'arrows', 'arrowstyle', 'arrowsize',
        'connectionstyle', 'label', 'min_source_margin', 'min_target_margin',
        # following kwargs are needed for estimating edge size
        'node_shape', 'node_size', 'nodelist'
    ],
    'edge_labels': [
        'ax', 'alpha', 'edge_labels', 'label_pos', 'font_size', 'font_color',
        'font_weight', 'font_family', 'bbox', 'clip_on', 'horizontalalignment',
        'verticalalignment'],
}
"""
Registered :mod:`nxgrph.draw_graph <tessif.visualize.nxgrph.draw_graph>`
key word arguments keyed by respective sub drawing utility name. Used for
filtering out incompatible key word arguments. Explicitly stated
here for easily maintained and expanded code.

Currently used exceptions:

.. execute_code::
    :hide_code:
    :hide_headers:
    :hide_output:

    from tessif.frused.paths import doc_dir
    import os
    from tessif.frused.defaults import nx_label_kwargs
    import pandas

    path = os.path.join(
         doc_dir, 'source', 'api', 'frused',
        'defaults', 'nx_label_kwargs.csv')

    df = pandas.DataFrame(data=nx_label_kwargs.values(),
                          index=nx_label_kwargs.keys())

    df.to_csv(path, header=None)

.. csv-table:: first column labels the keys
    :file: source/api/frused/defaults/nx_label_kwargs.csv
    :stub-columns: 1
"""


nx_label_kwargs = {
    'nodes': [
        'ax', 'nodelist', 'node_size', 'node_fill_size', 'node_color',
        'node_shape', 'alpha', 'cmap', 'vmin', 'vmax', 'linewidts',
        'edgecolors', 'label'],
    'labels': [
        'labels', 'font_size', 'font_color',  'font_family', 'font_weight',
        'alpha', 'horizontalalignment', 'verticalalignment', 'ax'],
    'edges': [
        'edglist', 'width', 'edge_color', 'style', 'alpha', 'edge_cmap',
        'edge_vmin', 'edge_vmax', 'ax', 'arrows', 'arrowstyle', 'arrowsize',
        'connectionstyle', 'label', 'min_source_margin', 'min_target_margin',
        # following kwargs are needed for estimating edge size
        'node_shape', 'node_size', 'nodelist'
    ],
    'edge_labels': [
        'ax', 'alpha', 'edge_labels', 'label_pos', 'font_size', 'font_color',
        'font_weight', 'font_family', 'bbox', 'clip_on', 'horizontalalignment',
        'verticalalignment'],
}
"""
Registered :mod:`nxgrph.draw_graph <tessif.visualize.nxgrph.draw_graph>`
key word arguments keyed by respective sub drawing utility name. Used for
filtering out incompatible key word arguments. Explicitly stated
here for easily maintained and expanded code.

Currently used exceptions:

.. execute_code::
    :hide_code:
    :hide_headers:
    :hide_output:

    from tessif.frused.paths import doc_dir
    import os
    from tessif.frused.defaults import nx_label_kwargs
    import pandas

    path = os.path.join(
         doc_dir, 'source', 'api', 'frused',
        'defaults', 'nx_label_kwargs.csv')

    df = pandas.DataFrame(data=nx_label_kwargs.values(),
                          index=nx_label_kwargs.keys())

    df.to_csv(path, header=None)

.. csv-table:: first column labels the keys
    :file: source/api/frused/defaults/nx_label_kwargs.csv
    :stub-columns: 1
"""

energy_system_nodes = {
    # Unique Identifiers
    'name': 'unspecified',
    'latitude': 0.0,
    'longitude': 0.0,
    'region': None,
    'sector': None,
    'carrier': None,
    'component': None,
    'node_type': None,
    'unspecified': 'Unspecified',

    # time serieses
    'timeseries': None,

    # singular values
    'accumulated_maximum': float('+inf'),
    'accumulated_minimum': 0.0,
    'active': 1,

    'characteristic_value': None,
    'costs_for_being_active': 0.0,

    'efficiency': 1.0,
    'emissions': 0.0,
    'exogenously_set': False,
    'exogenously_set_value': 0,
    'expandable': False,
    'fixed_expansion_ratios': True,
    'expansion_costs': 0.0,


    'flow_costs': 0.0,
    'final_soc': None,

    'gain_rate': 0.0,

    'initial_soc': 0.0,
    'initial_status': 1,
    'input': None,

    'interfaces': None,
    'installed_capacity': 0.0,

    'loss_rate': 0.0,

    'maximum': float('+inf'),
    'maximum_efficiency': 1.0,
    'maximum_expansion': float('+inf'),
    'maximum_flow_rate': float('+inf'),
    'maximum_shutdowns': float('+inf'),
    'maximum_startups': float('+inf'),

    'minimum': 0.0,
    'minimum_efficiency': 0.01,
    'minimum_expansion': 0.0,
    'minimum_flow_rate': 0.0,
    'minimum_downtime': 0,
    'minimum_uptime': 0,

    'negative_gradient': float('+inf'),
    'negative_gradient_costs': 0.0,

    'output': None,

    'positive_gradient': float('+inf'),
    'positive_gradient_costs': 0.0,

    'shutdown_costs': 0.0,
    'startup_costs': 0.0,
    'storage_capacity': 0.0,

    'variable_capacity': None,

    # chp values
    'chp_back_pressure': None,
    'chp_efficiency': {},
    'el_efficiency_wo_dist_heat': nts.MinMax(None, None),
    'enthalpy_loss': nts.MinMax(None, None),
    'min_condenser_load': None,
    'power_loss_index': None,
    'power_wo_dist_heat': nts.MinMax(None, None),

    # model specific
    'already_installed': 0.0,

    'back_pressure': False,

    'ideal': False,

    'milp': False,

    'number_of_connections': 1,

    'fine_region': 'Default Region'

}
"""
Fallback defaults for creating energy system nodes.

.. execute_code::
    :hide_code:
    :hide_headers:
    :hide_output:

    from tessif.frused.paths import doc_dir
    import os
    from tessif.frused.defaults import energy_system_nodes
    import dutils

    df = dutils.to_dataframe(energy_system_nodes, columns=3, fillvalue='')

    path = os.path.join(
        doc_dir, 'source', 'api', 'frused',
        'defaults', 'energy_system_nodes.csv')

    df.to_csv(path, header=None, index=None, na_rep='None')

.. csv-table::
    :file: source/api/frused/defaults/energy_system_nodes.csv
"""


registered_component_types = {
    'bus':
    ('bus',),
    'sink':
    ('sink', 'demand', 'export', 'excess',),
    'source':
    ('source', 'renewables', 'export', 'import', 'backup', 'commodity'),
    'storage':
    ('storage', 'generic_storage'),
    'transformer':
    ('transformer', 'mimo_transformer',),
    'connector':
    ('connector', 'link', 'line', 'connection',),
}
"""
Default energy system component identifiers and all types associated with them.

Used to :attr:`reorder <tessif.parse.reorder_esm>` an energy system mapping
during :mod:`parsing <tessif.parse>`.

Currently registered component types and their
:ref:`identifiers <Spellings_EnergySystemComponentIdentifiers>`.

.. execute_code::
    :hide_code:
    :hide_headers:
    :hide_output:

    from tessif.frused.paths import doc_dir
    import os
    from tessif.frused.defaults import registered_component_types as rct
    import pandas

    path = os.path.join(
        doc_dir, 'source', 'api', 'frused',
        'defaults', 'registered_component_types.csv')

    df = pandas.DataFrame(data=rct.values(),
                          index=rct.keys())

    df.to_csv(path, header=None)

.. csv-table:: first column labels the registration key
    :file: source/api/frused/defaults/registered_component_types.csv
    :stub-columns: 1
"""

addon_component_types = {
    'generic_chp': ('generic_chp',),
    'sito_flex_transformer': ('sito_flex_transformer',),
    'siso_nonlinear_transformer': (
        'offset_transformer', 'siso_nonlinear_transformer'),
}
"""
Added speciality energy system component identifiers and types associated with
them.

Currently registered component types and their
:ref:`identifiers <Spellings_EnergySystemComponentIdentifiers>` as added
components.

Note
----
Expand this dictionary when adding new speciality
:ref:`energy system components <Models_Tessif_Concept_ESC>` that are not quite
parseable using tessif's :ref:`approach <Models_Tessif_Concept>` but where a
specialised parsers exists or will be added. Like for example in
:meth:`tessif.transform.mapping2es.omf.generate_sito_flex_transformers`.


.. execute_code::
    :hide_code:
    :hide_headers:
    :hide_output:

    from tessif.frused.paths import doc_dir
    import os
    from tessif.frused.defaults import addon_component_types as act
    import pandas

    path = os.path.join(
        doc_dir, 'source', 'api', 'frused',
        'defaults', 'addon_component_types.csv')

    df = pandas.DataFrame(data=act.values(),
                          index=act.keys())

    df.to_csv(path, header=None)

.. csv-table:: first column labels the registration key
    :file: source/api/frused/defaults/addon_component_types.csv
    :stub-columns: 1
"""

registered_models = {
    'omf': spellings.oemof,
    'ppsa': spellings.pypsa,
    'fine': spellings.fine,
    'cllp': spellings.calliope,
}
"""
Registered energy supply system simulation models and their recognized
spelling variations.

.. execute_code::
    :hide_code:
    :hide_headers:
    :hide_output:

    from tessif.frused.paths import doc_dir
    import os
    from tessif.frused.defaults import registered_models as rm
    import pandas

    path = os.path.join(
        doc_dir, 'source', 'api', 'frused',
        'defaults', 'registered_models.csv')

    df = pandas.DataFrame(data=rm.values(),
                          index=rm.keys())

    df.to_csv(path, header=None)

.. csv-table:: first column labels the registration key
    :file: source/api/frused/defaults/registered_models.csv
    :stub-columns: 1


See Also
--------
:ref:`SupportedModels`
"""
# dash_visualize_defaults =



# tessif/frused/paths.py
# -*- coding: utf-8 -*-
"""
:mod:`~tessif.frused.paths` is a :mod:`tessif` subpackage aggregating
frequently used path implementations for conveniently accessing utilities and
data coming with tessif.
"""
import sys
import os
import pathlib
import ittools
import inspect


def find_subpath_incwd(path, includes=['tessif', ], excludes=['.git', ]):
    """ Function to find location of path. Credit to:
    https://stackoverflow.com/a/41546830

    Note
    -----
    To make this utility work os independently it is advised to supply
    path using :func:`os.path.join`.

    Parameters
    ----------
    path : str
        String representation of the directory/file to find the path for.
        Supplying this paremter us ing :func:`os.path.join` makes this
        utility platform independent
    includes : :class:`~collections.abc.Iterable`, str
        String or iterable of strings that should be part of the
        found path.
    exludes : :class:`~collections.abc.Iterable`, str
        String or iterable of strings that should **not** be part of the
        found path.

    Return
    ------
    found_paths : list
        List of found paths
    """

    excludes = ittools.itrify(excludes)
    if getattr(sys, 'frozen', False):
        # The application is frozen
        datadir = os.path.dirname(sys.executable)

    else:
        # The application is not frozen
        # Change this bit to match where you store your data files:
        datadir = os.path.dirname(os.getcwd())

    candidates = list(
        pathlib.Path(pathlib.Path(datadir).parent).glob(
            os.path.join('**', path)))

    found_paths = [candidate.as_posix() for candidate in candidates
                   if not any(exclude in candidate.as_posix()
                              for exclude in excludes)
                   and all(include in candidate.as_posix()
                           for include in includes)]
    return found_paths


root_dir = os.path.normpath(os.path.join(inspect.getfile(
    inspect.currentframe()).split('paths')[0], '..', ))
""" Tessif's root directory."""

#: Tessif's documentation directory
doc_dir = example_dir = os.path.normpath(
    os.path.join(root_dir, '..', '..', 'docs'))

log_dir = os.path.join(root_dir, 'write', 'logs')
""" Tessif's log directory. """

# Make sure the logging path exists.
pathlib.Path(log_dir).mkdir(parents=True, exist_ok=True)

example_dir = os.path.join(root_dir, 'examples')
""" Tessif's example directory """

write_dir = os.path.join(root_dir, 'write')
""" Tessif's write (output) directory"""

#: Tessif's tests directory
tests_dir = os.path.normpath(
    os.path.join(root_dir, '..', '..', 'tests'))



# src/tessif/frused/utils.py
"""Tessif (random) utility collection."""


def greyscale2hex(greyscale, minn=0.0, maxn=1.0):
    """Correlate a number within a certain range to a hex encoded gray value.

    Parameters
    ----------
    greyscale: ~numbers.Number
        Number between :paramref:`minn` and :paramref:`maxn` representing the
        greyscale. Where :paramref:`minn` =  ``'white'`` and
        :paramref:`maxn` = ``'black'``.
    minn: ~number.Number, default=0.0
        Lower boundary resulting in a white color hex value
    maxn: ~number.Number, default=1.0
        Upper boundary resulting in a black color hex value

    Returns
    -------
    hex
        single hex color value representing the correlated grayscale color.

    Examples
    --------
    >>> greyscale2hex(.3)
    '#b2b2b2'
    >>> greyscale2hex(.7)
    '#4c4c4c'
    >>> greyscale2hex(50, 0, 100)
    '#7f7f7f'
    >>> greyscale2hex(1e10, 0, 1e20)
    '#fefefe'
    >>> greyscale2hex(0)
    '#ffffff'
    >>> greyscale2hex(1)
    '#000000'
    >>> greyscale2hex(0.0)
    '#ffffff'
    >>> greyscale2hex(1.0)
    '#000000'
    """
    greyscale = _clamp(greyscale, minn, maxn)
    rgb_int = int((maxn - greyscale) / maxn * 255)
    return f"#{rgb_int:02x}{rgb_int:02x}{rgb_int:02x}"


def _clamp(number, minn, maxn):
    """Clamp number to be within [minn, maxn]."""
    return max(min(maxn, number), minn)



# src/tessif/namedtuples.py
# pylint: disable=invalid-name
"""Dummy named tuples file to enebale tessif-examples."""
import collections
import typing

node_uid_seperator = "_"
node_uid_style = "name"


class UidBase(typing.NamedTuple):

    name: str
    latitude: float
    longitude: float
    region: str
    sector: str
    carrier: str
    component: str
    node_type: str

    def __str__(self):
        return node_uid_seperator.join(
            str(getattr(self, i)) for i in node_uid_styles[
                node_uid_style])


class Uid(UidBase):
    r"""
    **U**\ nique **ID**\ entifiers for each energy system component.

    Essential part of :ref:`tessif's labeling concept <Labeling_Concept>`.

    For an energy system component to be unique, the compination of the
    :ref:`uid components <Namedtuples_UidComponents>` beeing used must be
    unique. Which combination is used internally depends on on the used
    :attr:`node_uid_styles` which can be selected using
    :attr:`configurations.node_uid_style
    <tessif.frused.configurations.node_uid_style>`.

    :class:`~typing.NamedTuple` containing meta and geospatial information.

    Note
    ----
    Used for internal results mapping. See
    :attr:`~tessif.frused.configurations.node_uid_style` and
    :attr:`node_uid_styles` for more details.
    Especially when having issues with non-unique uid solver issues.

    Note
    ----
    A UID's components are the :ref:`Parameters <Namedtuples_UidComponents>`
    of this class. For recognized variations of these during data parsing
    see also: :ref:`Spellings_UidComponents`


    .. _Namedtuples_UidComponents:

    Parameters
    ----------
    name: str
        Identifier replacing the original label.
        i.e: 'PV', 'Gas_CHP', ...

    latitude: :class:`~numbers.Number`, default None
        Geospatial latitude in degree

    longitude: :class:`~numbers.Number`, default None
        Geospatial longitude in degree

    region: str, default None
        Arbitrary regional categorization string.
        'Europe', 'global', ...

    sector: str, default None
        Arbitrary sector categorization string.

    carrier: str, default None
        Arbitrary energy carrier categorization string.

    component: str, default None
        One of the :ref:`energy system component identifiers
        <Models_Tessif_Concept_ESC>`.

    node_type: str, default None
        Arbitrary node type categorization string.
        i.e. "Combined_Cycle", "Renewable", ...

    warning
    -------
    Uid component defaults have to be set manually, according to
    :attr:`defaults.energy_system_nodes
    <tessif.frused.defaults.energy_system_nodes>` to avoid circular imports!

    return
    ------
    uid: ~typing.NamedTuple
        Namedtuple instance object serving as uid
    """

    def __new__(cls, name,
                latitude=None,
                longitude=None,
                region=None,
                sector=None,
                carrier=None,
                component=None,
                node_type=None):
        self = super().__new__(cls, name,  latitude, longitude, region,
                               sector, carrier, component, node_type)

        return self

    @classmethod
    def reconstruct(cls, string_representation):
        """Reconstruct a :class:`Uid` from its :ref:`string representation
        <Labeling_Concept>`."""
        deconstructed_uid = string_representation.split(
            config.node_uid_seperator)
        dict_representation = dict()
        for pos, attr in enumerate(node_uid_styles[config.node_uid_style]):
            if pos > len(deconstructed_uid) - 1:
                msg = (
                    "Index out of range\n" +
                    "The node_uid_style specificiation needs to be done " +
                    "BEFORE constructing a Uid implying it also needs to be " +
                    "Done BEFORE CREATING a model specific energy system\n" +
                    "(Use 'qualname' before constructing the model specific " +
                    "energy system and use more specified styles before " +
                    "attempting a Uid reconstruction.)"
                )
                raise IndexError(msg)

            dict_representation[attr] = deconstructed_uid[pos]

        return cls(**dict_representation)


node_uid_styles = {
    'name': ['name'],
    'qualname': [i for i in Uid.__new__.__code__.co_varnames
                 if i not in ['self', 'cls']],
    'coords': ['name', 'latitude', 'longitude'],
    'region': ['name', 'region'],
    'sector': ['name', 'sector'],
    'carrier': ['name', 'carrier'],
    'component': ['name', 'component'],
    'node_type': ['name', 'node_type']}
"""
Provide possible internal node label representation styles. Use
:attr:`configurations.node_uid_style
<tessif.frused.configurations.node_uid_style>`. For selecting a particular
style.

Currently supported styles are (first column identifies the mapping key):

    .. execute_code::
        :hide_code:
        :hide_headers:
        :hide_output:

        import tessif.frused.namedtuples as nts
        from tessif.frused.paths import doc_dir
        import pandas as pd
        import os

        df = pd.DataFrame(
            index=nts.node_uid_styles.keys(),
            data=nts.node_uid_styles.values())

        path = os.path.join(
            doc_dir, 'source', 'api', 'frused',
            'namedtuples', 'node_uid_styles.csv')

        df.to_csv(path, header=None)

    .. csv-table::
        :file: source/api/frused/namedtuples/node_uid_styles.csv
        :stub-columns: 1

Combined with :attr:`~tessif.frused.configurations.node_uid_seperator`
a node of::

    lbl = tessif.frused.namedtuples.uid(
        name='test',
        latitude='lat',
        longitude='long',
        region='hamburg',
        sector='power',
        carrier='wind',
        component='source',
        node_type='renewable')

results into following representaion, depending on the mapping used:

    .. execute_code::
        :hide_code:
        :hide_headers:

        import tessif.frused.namedtuples as nts
        import tessif.frused.configurations as configs

        lbl = nts.Uid(
            name='test',
            latitude='lat',
            longitude='long',
            region='hamburg',
            sector='power',
            carrier='wind',
            component='source',
            node_type='renewable')

        nls = nts.node_uid_styles

        for key in nls.keys():
            configs.node_uid_style = key
            print(configs.node_uid_style, '->', lbl)
"""

MinMax = collections.namedtuple("MinMax", ["min", "max"])
"""
Corespondent minimum and maximum value pair. Usually used for energy system
value parings like flow limits (i.e: minimum and maximum flow boundaries).
"""

OnOff = collections.namedtuple("OnOff", ["on", "off"])
"""
Corespondent value pair for describing parameters dependent on a boolean
status. Usually used for energy system value parings describing nonconvex
behaviour like i.e: minimum uptime and downtime.
"""

InOut = collections.namedtuple("InOut", ["inflow", "outflow"])
"""
Corespondent inflow outflow value pair. Usually used for energy system value
parings describing different behaviours of the same component depending on in
or outflow like i.e: Storage efficiency
"""

PositiveNegative = collections.namedtuple(
    "PositiveNegative",
    ["positive", "negative"],
)
"""
Corespondent value pair for describing parameters dependent on directions
expressed as positive or negative. Usually used for energy system value
parings describing changes between timesteps like i.e: power production
gradients.
"""

NodeColorGroupings = collections.namedtuple(
    'NodeColorGroupings',
    ['component', 'name', 'carrier', 'sector'])
"""
Node color groups mapped by format like
:class:`~tessif.transform.es2mapping.ESTransformer` objects.

Mapping colors are defined in :mod:`tessif.frused.themes`.
Unique energy system component identifier (:class:`uid`) parts are specified
in :attr:`~tessif.frused.spellings.energy_system_component_identifiers`.

Parameters
----------

component: defaultdict(list)
    Dictionairy of node colors with component searched for in
    :attr:`~tessif.frused.spellings.energy_system_component_identifiers`

label : defaultdict(list)
    Dictionairy of node colors with tags searched for in str(node.label)

carrier : defaultdict(list)
    Dictionairy of node colors with tags searched for in node.label.carrier

sector : defaultdict(list)
    Dictionairy of node colors with tags searched for in node.label.sector
"""



# src/tessif/cli.py
"""Module providing the Tessif Command Line Interface (CLI)."""
import importlib

import click


@click.group()
def main_cli_entry():
    """Tessif-Command main entry point."""
    pass


@main_cli_entry.command()
def greet():
    """Greet the world."""
    click.echo("Hello World!")


@main_cli_entry.command()
@click.option(
    "--system_model_location",
    help="Tessif System-Model Location",
)
@click.argument("plugin")
def tropp(system_model_location, plugin):
    """TRansform Optimize and Post-Process."""
    click.echo("TRansform Optimize and Post-Process!\n")

    module_name = plugin.replace("-", "_")
    plugin_module = importlib.import_module(module_name)

    import oemof.solph

    click.echo("Using Plugin:")
    click.echo(f"{plugin} version {plugin_module.__version__}\n")
    click.echo("Using the modelling suit:")
    click.echo(f"oemof.solph version {oemof.solph.__version__}\n")

    click.echo("On System Model Stored In:")
    click.echo(system_model_location)
    click.echo()

    from tessif.model.energy_system import AbstractEnergySystem

    restored_es = AbstractEnergySystem(uid="This Instance Is Restored")
    msg = restored_es.unpickle(system_model_location)
    click.echo("Succesfully deserialized tessif system model!")
    click.echo("Nodes present in the tessif system model:")
    click.echo([node.uid.name for node in restored_es.nodes])
    click.echo()

    # t2o = importlib.import_module(f"{module_name}.tsf2omf")
    tool_system_model = plugin_module.transform(restored_es)

    click.echo("Nodes present in the oemof system model:")
    click.echo([str(node) for node in tool_system_model.nodes])

    optimized_system_model = plugin_module.optimize(tool_system_model)

    click.echo("Optimization Succesfull!")
    click.echo(f"Global Costs: {optimized_system_model.results['global']['costs']}")


# if __name__ == '__main__':
#     main_cli_entry()



# tessif/frused/configurations
# -*- coding: utf-8 -*-
"""
:mod:`~tessif.frused.configurations` is a :mod:`tessif` subpackage aggregating
simulation parameters, used naming and unit conventions as well as some
logging behavior.

It serves as main reference point for adjusting tessif's simulative and
parsing behavior.
"""
import os
from tessif.frused.paths import log_dir

temporal_resolution = 'hourly'
"""
Currently used temporal resolution. Must be one of the keys found in
:attr:`~tessif.frused.resolutions.temporals`.

Currently set to:

    .. execute_code::
        :hide_code:
        :hide_headers:

        import tessif.frused.configurations as config
        print(config.temporal_resolution)
"""

node_uid_style = 'name'
"""
Switch for tweaking internal node uid representation style.

Useful for conveniently changing internal mapping behaviour. Must be one of
:attr:`~tessif.frused.namedtuples.node_uid_styles`.

Somehting like ``name`` allows for quick and intuitive node accessing while
sacrificing the possibility of 2 nodes having the same
:paramref:`~tessif.frused.namedtuples.uid.name`.

``qualname`` on the other hand maps everything to the fully qualified name.
Meaning only ever the full combination of all
:class:`~tessif.frused.namedtuples.uid` attributes has to be unique per
energy system.

.. warning::

    Tessif's doctests assumes ``node_uid_style = 'name'`` which is the most
    basic and intuitive way of mapping nodes. Designed for the use of
    relatively small energy systems (what ever that means).

For a list of available styles and their key (the string set to
:attr:`node_uid_style`) see :attr:`tessif.frused.namedtuples.node_uid_styles`.

Currently set to:

    .. execute_code::
        :hide_code:
        :hide_headers:

        import tessif.frused.configurations as config
        print(config.node_uid_style)
"""

node_uid_seperator = '_'
"""
Seperate different tags of the same
:attr:`node uid <tessif.frused.namedtuples.uid>`

Seperate symbol for (uniquely) identifying a node's uid using various tags of
the :attr:`namedtuples implementation <tessif.frused.namedtuples.uid>`.

Currently set to:

    .. execute_code::
        :hide_code:
        :hide_headers:

        import tessif.frused.configurations as config
        print(config.node_uid_seperator)

So ``TAG1{NODE_UID_SEPERATOR}TAG2`` results in:

    .. execute_code::
        :hide_code:
        :hide_headers:

        import tessif.frused.configurations as config
        print('TAG1{}TAG2'.format(config.node_uid_seperator))
"""

timeseries_seperator = '.'
"""
Seperate energy system object and timeseries value.

Serperator symbol for identifying energy system object and its timeseries
values when reading in data.

Standard syntax::

    {ES_OBJECT}{SEPERATOR}{TIMESERIES_PARAMETER}.

Currently set to:

    .. execute_code::
        :hide_code:
        :hide_headers:

        import tessif.frused.configurations as config
        print(config.timeseries_seperator)

So ``PV{SEPERATOR}max`` and ``Onshore{SEPERATOR}fix`` results in:

    .. execute_code::
        :hide_code:
        :hide_headers:

        import tessif.frused.configurations as config
        print('PV{}max'.format(config.timeseries_seperator))
        print('Onshore{}fix'.format(config.timeseries_seperator))
"""

mimos = 10
"""
Number of seperate inputs/outputs supported for multiple input output energy
system transformers.

Currently set to:

    .. execute_code::
        :hide_code:
        :hide_headers:

        import tessif.frused.configurations as config
        print(config.mimos)
"""

power_reference_unit = 'MW'
"""
Unit to display power results with.

Currently set to:

    .. execute_code::
        :hide_code:
        :hide_headers:

        import tessif.frused.configurations as config
        print(config.power_reference_unit)
"""

cost_unit = '€'
"""
Unit representing the costs.

Currently set to:

    .. execute_code::
        :hide_code:
        :hide_headers:

        import tessif.frused.configurations as config
        print(config.cost_unit)
"""

#: Dictionairy configuring tessif's logging locations and file names
logging_file_paths = {
    'debug': os.path.join(log_dir, 'debug.log'),
    'content': os.path.join(log_dir, 'content.log'),
    'timings': os.path.join(log_dir, 'timings.log'),
}
"""
Tessif's logging locations and filenames.

Currently set to:

    .. execute_code::
        :hide_code:
        :hide_headers:

        from tessif.frused.configurations import logging_file_paths as lgn
        print(lgn['debug'])
        print(lgn['content'])
        print(lgn['timings'])

Warning
-------
Logging root directory must exist for the logging file location configuration
to work as expected.
"""

spellings_logging_level = 'warning'
"""
`logging level
<https://docs.python.org/3/library/logging.html#logging-levels>`_
used by :meth:`spellings.get_from <tessif.frused.spellings.get_from>`.

Must be one of the keys found in :attr:`~tessif.write.log.logging_levels`.

Currently set to:

    .. execute_code::
        :hide_code:
        :hide_headers:

        import tessif.frused.configurations as config
        print(config.spellings_logging_level)
"""
